/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation version 2.1.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
 ***************************************************************************/
#ifndef CONEXUSDATA_H
#define CONEXUSDATA_H

#include <string>
#include <boost/shared_ptr.hpp>

namespace Conexus {

/**
 * This structure represents a data block allocated via a smart pointer.
 *
 * Additional it provides several typedefs that may be used to ensure type consistency both within
 * the conexus library, and with applications using the conexus library.
 *
 * @author Rick L Vinyard Jr
 * @ingroup conexus
 */
struct Data {
    /**
     * @typedef Octet
     * Typedefs the concept of an octet as an unsigned 8 bit integer.
     */
    typedef uint8_t Octet;

    /**
     * @typedef Octets
     * Provides a convenience typedef of a smart pointer to data.
     * Hint: access externally as Data::Octets
     */
    typedef boost::shared_ptr<Octet> Octets;
    
    typedef boost::shared_ptr<const Octet> ConstOctets;
    
    /**
     * Default constructor initializes data size to zero and relies on smart pointer constructor to
     * initialize pointer to NULL.
     */
    Data(): m_size(0) { }

    /**
     * Constructor which accepts a pointer to data and size and creates a duplicate
     */
    Data(const void* d, size_t s): m_size(s) {
        m_data = Octets(new Octet[m_size], boost::checked_array_deleter<Octet>() );
        memcpy(m_data.get(), d, m_size);
    }

    /**
     * Constructor which accepts a smart pointer to data and data size in octets.
     */
    Data(Octets d, size_t s): m_data(d), m_size(s) { }

    /**
     * Constructor which creates a data object of a given size in octets.
     */
    Data(size_t s): m_size(s) {
      m_data = Octets(new Octet[s], boost::checked_array_deleter<Octet>() );
    }

    Octets data() { return m_data; }
    
    ConstOctets data() const { return boost::static_pointer_cast<const Octet>(m_data); }
    
    void set_data( Octet const* newdata, size_t newsize ) {
      m_data = Octets( new Octet[newsize], boost::checked_array_deleter<Octet>() );
      memcpy( m_data.get(), newdata, newsize );
      m_size = newsize;
    }
    
    size_t size() const { return m_size; }
    
    void set_size( size_t s ) {
      if ( m_size == s )
        return;
      Octets olddata = m_data;
      m_data = Octets( new Octet[s], boost::checked_array_deleter<Octet>() );
      memcpy( m_data.get(), olddata.get(), ((m_size<s)?m_size:s) );
      m_size = s;
    }
    
    /**
     * Performs a deep copy of this data object.
     * Assignment results in a shallow copy of this data object, which is in accordance with
     * the semantics of the smart pointers used. However, at times a deep copy may be necessary.
     * This method allocates a new memory chunk, copies the existing data into that data chunk,
     * and returns an object containing a smart pointer to the underlying data.
     */
    Data clone() const {
      Data retval(m_data.get(), m_size);
      return retval;
    }

    /**
     * Overloaded dereference operator provides direct access to actual data. If you want access
     * to the smart pointer, use the @em data member instead.
     */
    operator Octet*() {
        return m_data.get();
    }
    
    operator Octet const*() const {
      return static_cast<Octet const*>(m_data.get());
    }
    
    Octet* raw_data() {
      return m_data.get();
    }
    
    Octet const* raw_data() const {
      return static_cast<Octet const*>(m_data.get() );
    }

    /** Returns a hex string representation of the data */
    std::string hex_string(std::string separator=std::string() ) const
    {
      std::string s;
      size_t i;
      uint8_t* p;
      char hex[3];

      hex[2] = 0x00;

      p = m_data.get();
      for (i=0; i < m_size; i++)
      {
        if (i != 0)
          s += separator;
        hex[0] = '0' + (*p >> 4);
        hex[1] = '0' + (*p & 0x0F);
        if (hex[0] > '9')
          hex[0] += 7;
        if (hex[1] > '9')
          hex[1] += 7;
        s += hex;
        p++;
      }
      return s;
    }

    /** Frees any allocated memory and resets size to 0 */
    void clear() {
      m_data.reset();
      m_size = 0;
    };
    
  protected:

    /**
     * Smart pointer to the underlying data.
     */
    Octets m_data;

    /**
     * Size of the data in bytes.
     */
    size_t m_size;
};


}

#endif
