/***************************************************************************
*   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
*   rvinyard@cs.nmsu.edu                                                  *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU Lesser General Public License as        *
*   published by the Free Software Foundation version 2.1.                *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU Lesser General Public      *
*   License along with this library; if not, write to the                 *
*   Free Software Foundation, Inc.,                                       *
*   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
***************************************************************************/
#include "dataserver.h"

#include <conexus/utility.h>

namespace Conexus
{

  DataServer::DataServer() :
      m_terminate(false),
      m_dispatcher( NULL ),
      m_use_dispatcher( false )
  {
    m_running[ READ ] = false;
    m_running[ WRITE ] = false;
    m_timeval.tv_sec = 0;
    m_timeval.tv_usec = 500000;
  }


  DataServer::~DataServer()
  {
    if ( this->is_running() )
      stop();
  }
  
  void DataServer::set_responsiveness(long r)
  {
    m_timeval.tv_usec = r;
  }

  long DataServer::responsiveness()
  {
    return m_timeval.tv_usec;
  }

  bool DataServer::is_running()
  {
    return m_running[ READ ] | m_running[ WRITE ];
  }

  bool DataServer::is_read_running()
  {
    return m_running[ READ ];
  }

  bool DataServer::is_write_running()
  {
    return m_running[ WRITE ];
  }

  void DataServer::start( ReadWrite rw )
  {
    this->start( rw, use_dispatcher_default() );
  }
  
  void DataServer::start( ReadWrite rw, bool use_dispatcher )
  {
    // Start READ first then WRITE
    for ( int i = READ; i <= WRITE; i++ ) {
      if ( ! ( rw & i ) )
        continue;

      m_mutex_can_start[ i ].lock();

      if ( m_running[ i ] ) {
        m_mutex_can_start[ i ].unlock();
        continue;
      }

      if ( i == READ ) {
        m_use_dispatcher = use_dispatcher;
        if ( m_use_dispatcher ) {
          m_dispatcher = new Glib::Dispatcher();
          if ( m_dispatcher && ! m_dispatcher_connection )
            m_dispatcher_connection = m_dispatcher->connect( sigc::mem_fun( *this, &DataServer::emit_received_data ) );
        }
      }

      m_running[ i ] = true;

      if ( i == READ )
        m_thread[ i ] = Glib::Thread::create( sigc::mem_fun(*this, &DataServer::read_thread_main), true );
      else
        m_thread[ i ] = Glib::Thread::create( sigc::mem_fun(*this, &DataServer::write_thread_main), true );

      m_mutex_can_start[ i ].unlock();
    }
  }

  void DataServer::stop( ReadWrite rw )
  {
    // Stop READ first then WRITE
    for ( int i = READ; i <= WRITE; i++ )
      if ( ( rw & i ) && m_running[ i ] ) {
        m_mutex_can_start[i].lock(); // Grab a start lock to prevent starting before cleanup is done
        m_terminate = true;
        m_thread[i]->join(); // Wait for the cancel to take effect before unlocking
        m_running[ i ] = false; // We're stopped and can't start
        if ( i == READ )
          this->read_thread_cleanup();
        else
          this->write_thread_cleanup();
        m_mutex_can_start[i].unlock();
      }
  }

  void DataServer::read_thread_cleanup()
  {
    if ( m_dispatcher_connection )
      m_dispatcher_connection.disconnect();
    if ( m_dispatcher ) {
      m_use_dispatcher = false;
      delete m_dispatcher;
      m_dispatcher = NULL;
    }
  }

  void DataServer::write_thread_cleanup()
  {}

  sigc::signal<void, const Data&> DataServer::signal_data()
  {
    return m_signal_data;
  }

  void DataServer::read_thread_main()
  {}

  void DataServer::write_thread_main()
  {}

  void DataServer::queue_received_data( const Data& d )
  {
    m_mutex_queue[READ].lock();
    m_read_queue.push( d );
    m_mutex_queue[READ].unlock();
    if ( m_use_dispatcher )
      m_dispatcher->emit();
    else
      this->emit_received_data();
  }

  void DataServer::queue_write_data( const Data& d )
  {
    m_mutex_queue[WRITE].lock();
    m_write_queue.push( d );
    m_mutex_queue[WRITE].unlock();
  }

  void DataServer::emit_received_data( )
  {
    Data current;
    m_mutex_queue[READ].lock();
    while ( ! m_read_queue.empty() ) {
      current = m_read_queue.front();
      m_read_queue.pop();
      m_mutex_queue[READ].unlock();
      m_signal_data.emit( current );
      m_signal_data_received.emit( current.size() );
      m_mutex_queue[READ].lock();
    }
    m_mutex_queue[READ].unlock();
  }

  sigc::signal< void, size_t > DataServer::signal_data_received( )
  {
    return m_signal_data_received;
  }

  sigc::signal< void, size_t > DataServer::signal_data_transmitted( )
  {
    return m_signal_data_transmitted;
  }

}
