/***************************************************************************
*   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
*   rvinyard@cs.nmsu.edu                                                  *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU Lesser General Public License as        *
*   published by the Free Software Foundation version 2.1.                *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU Lesser General Public      *
*   License along with this library; if not, write to the                 *
*   Free Software Foundation, Inc.,                                       *
*   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
***************************************************************************/
#ifndef CONEXUSDATASERVER_H
#define CONEXUSDATASERVER_H

#include <set>
#include <queue>

#include <sys/select.h>

#include <glibmm/dispatcher.h>
#include <glibmm/thread.h>

#include <conexus/enums.h>
#include <conexus/object.h>
#include <conexus/data.h>

namespace Conexus {

  /**
   * @author Rick L Vinyard Jr <rvinyard@cs.nmsu.edu>
   */
  class DataServer : public Object {
    public:
      DataServer();

      virtual ~DataServer();

      /**
       * Start the endpoint. If the endpoint is configured for multi-threaded read/write this
       * will create and start the threads.
       * 
       * Whether a dispatcher is used or not is dependent upon the global setting which defaults
       * to false, but may be set with Conexus::init() or Conexus::set_use_dispatcher_default().
       * 
       * If you want to handle it manually, endpoints are capable of reading/writing without
       * ever calling start/stop.
       * 
       * @param rw Determines which aspects of this endpoint are threaded. If set to READ, then
       * read operations will be handled in a separate thread. If set to WRITE, writes will be
       * handled in a separate thread, and if set to READWRITE then reads will be handled in one
       * thread while writes are handled in another thread.
       */
      virtual void start( ReadWrite rw = READ );
      
      /**
       * Starts the endpoint, and forces it to use a specific dispatcher value possibly overriding
       * the global default.
       */
      virtual void start( ReadWrite rw, bool use_dispatcher );
      
      /**
       * Stops any running threads. Stopping is not immediate. If a read operation is being performed
       * the current read will complete before the thread is stopped.
       */
      virtual void stop( ReadWrite rw = READ );

      /**
       * Signal emitted when data is read. Connect to this signal if you want to receive all
       * data read by the endpoint. The data object received is unique for each callback
       * point, allowing a callback to modify the data set without effect subsequent callbacks.
       */
      sigc::signal<void, const Data&> signal_data();
      
      /**
       * Signal emitted when data is received notifying the callback of the number of bytes read.
       * Connect to this signal if you need to know when data is received, and how much, but don't
       * want the data contents.
       */
      sigc::signal<void, size_t> signal_data_received();
      
      /**
       * Signal emitted when data is transmitted, notifying the callback of the number of bytes transmitted.
       * Connect to this signal if you need to know when data is transmitted, and how much, but don't
       * want the transmitted data contents.
       */
      sigc::signal<void, size_t> signal_data_transmitted();

      /**
       * Sets the responsiveness to add_*, remove_* and stop commands in microseconds;
       * lower is more responsive; larger is less responsive; default is 500000 microseconds (or 500 ms).
       *
       * In general this refers to the timeout period of the @em select() call used internally.
       * Each cycle of the server I/O thread acquires several mutex locks at the beginning
       * of the cycle and releases them at the end of the cycle.
       *
       * This value shortens the cycle period, thereby making the server more responsive to commands at
       * the cost of increasing overhead.
       */
      virtual void set_responsiveness(long r);

      /**
       * Returns the server responsiveness level.
       */
      virtual long responsiveness();

      /** True if the endpoint is running with threaded reads and/or writes */
      virtual bool is_running();
      
      /** True if the endpoint is running with threaded reads */
      virtual bool is_read_running();
      
      /** True if the endpoint is running with threaded writes */
      virtual bool is_write_running();

      /** Conexus library string type identifier */
      virtual const std::string& object_type() { static std::string s("Conexus::DataServer"); return s; }

    protected:
      bool m_running[ 2 ];
      Glib::Thread* m_thread[ 2 ];
      bool m_terminate;
      struct timeval m_timeval;

      Glib::Mutex m_mutex_queue[ 2 ];
      Glib::Mutex m_mutex_can_start[ 2 ];

      Glib::Dispatcher* m_dispatcher;
      bool m_use_dispatcher;
      sigc::connection m_dispatcher_connection;

      sigc::signal<void, const Data&> m_signal_data;
      sigc::signal<void, size_t> m_signal_data_received;
      sigc::signal<void, size_t> m_signal_data_transmitted;
      std::queue<Data> m_read_queue;
      std::queue<Data> m_write_queue;

      virtual void read_thread_main();
      virtual void read_thread_cleanup();
      virtual void write_thread_main();
      virtual void write_thread_cleanup();

      void queue_received_data( const Data& d );
      void queue_write_data( const Data& d );
      void emit_received_data( );

  };

}

#endif
